/*
 * Support for executing code in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>
#include <l4/sys/factory.h>
#include <l4/sys/task.h>
#include <l4/sys/thread.h>
#include <l4/util/elf.h>
#include <l4/util/util.h>

#include <l4/libloader/adjust_stack>

#include <fsclient/client.h>
#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>
#include <ipc/server.h>
#include <mem/memory_utils.h>
#include <systypes/fcntl.h>

#include <map>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "pager_object_interface.h"
#include "pager_object_server.h"



/* UTCB properties.
   See: moe/server/src/loader.cc */

enum ipc_exec_utcb
{
  Default_max_threads = 16,
#ifdef ARCH_mips
  Utcb_area_start = 0x73000000,
#else
  Utcb_area_start = 0xb3000000,
#endif
};

static l4_fpage_t get_utcb_fpage()
{
  /* UTCB location and size. */

  int utcb_log2size = page_order(Default_max_threads * L4_UTCB_OFFSET);

  /* Round up to at least one page. */

  if (utcb_log2size < L4_PAGESHIFT)
    utcb_log2size = L4_PAGESHIFT;

  return l4_fpage(Utcb_area_start, utcb_log2size, 0);
}



/* Mapped region data structures. */

class MappedRegion
{
public:
  l4_addr_t start;
  unsigned int log2size;
  l4_umword_t flags;
  l4_addr_t map_start;

  explicit MappedRegion()
  : start(0), log2size(0), flags(0), map_start(0)
  {
  }

  explicit MappedRegion(l4_addr_t start, unsigned int log2size,
                        l4_umword_t flags, l4_addr_t map_start)
  : start(start), log2size(log2size), flags(flags), map_start(map_start)
  {
  }
};



/* Program segment abstraction. */

class Segment
{
protected:
  MappedRegion _region;

public:
  /* Allocated memory. */

  char *buf;
  l4re_ds_t ds;

  /* Segment base and corresponding region base. */

  l4_addr_t base, region_base;

  /* Segment size and corresponding region size. */

  offset_t size, region_size;

  /* Offset of segment content within the region. */

  offset_t region_offset;

  /* Access flags. */

  l4re_rm_flags_t flags;

  /* File access details. */

  offset_t file_offset, file_contents;

  explicit Segment(offset_t base, offset_t size, l4re_rm_flags_t flags,
                   offset_t file_offset = 0, offset_t file_contents = 0)
  : base(base), size(size), flags(flags), file_offset(file_offset), file_contents(file_contents)
  {
    region_base = trunc(base, L4_PAGESIZE);
    region_offset = base - region_base;
    region_size = round(size, L4_PAGESIZE);
  }

  long allocate();

  long fill(file_t *file);

  MappedRegion &region();

  l4_addr_t region_address(char *address);

  l4_addr_t region_address(l4_addr_t address);
};

long Segment::allocate()
{
  return ipc_allocate_align(size, L4RE_RM_F_SEARCH_ADDR | flags,
                           L4_PAGESHIFT, (void **) &buf, &ds);
}

long Segment::fill(file_t *file)
{
  if (!file_contents)
    return L4_EOK;

  memset(buf, 0, region_size);
  client_seek(file, file_offset, SEEK_SET);
  offset_t nread = client_read(file, buf + region_offset, file_contents);

  if (nread < file_contents)
    return -L4_EIO;
  else
    return L4_EOK;
}

MappedRegion &Segment::region()
{
  _region = MappedRegion((l4_addr_t) buf, page_order(region_size), flags, region_base);
  return _region;
}

l4_addr_t Segment::region_address(char *address)
{
  return (l4_addr_t) ((address - buf) + (char *) region_base);
}

l4_addr_t Segment::region_address(l4_addr_t address)
{
  return (address - (l4_addr_t) buf) + region_base;
}



/* A simple system pager also acting as a region mapper. */

typedef std::map<l4_addr_t, MappedRegion> MappedRegions;

class ExecPager : public PagerObject
{
protected:
  MappedRegions _regions;

public:
  virtual void add(MappedRegion region)
  {
    _regions[region.map_start] = region;
  }

  /* Notification methods. */

  virtual long exception(l4_exc_regs_t regs,
                         l4_snd_fpage_t *region);

  virtual long page_fault(l4_umword_t pfa, l4_umword_t pc,
                          l4_snd_fpage_t *region);

  /* Region manager/mapper methods. */

  virtual long attach(address_t *start, offset_t size, map_flags_t flags,
                      l4_cap_idx_t ds, address_t offset, unsigned char align);

};

/* Handle a general exception. */

long ExecPager::exception(l4_exc_regs_t regs, l4_snd_fpage_t *region)
{
  (void) region;

  printf("exception(...) -> pfa = %lx, pc = %lx\n", l4_utcb_exc_pfa(&regs), l4_utcb_exc_pc(&regs));

  printf("r15 = %lx\n", regs.r15);
  printf("r14 = %lx\n", regs.r14);
  printf("r13 = %lx\n", regs.r13);
  printf("r12 = %lx\n", regs.r12);
  printf("r11 = %lx\n", regs.r11);
  printf("r10 = %lx\n", regs.r10);
  printf("r9 = %lx\n", regs.r9);
  printf("r8 = %lx\n", regs.r8);
  printf("rdi = %lx\n", regs.rdi);
  printf("rsi = %lx\n", regs.rsi);
  printf("rbp = %lx\n", regs.rbp);
  printf("pfa = %lx\n", regs.pfa);
  printf("rbx = %lx\n", regs.rbx);
  printf("rdx = %lx\n", regs.rdx);
  printf("rcx = %lx\n", regs.rcx);
  printf("rax = %lx\n", regs.rax);
  printf("trapno = %lx\n", regs.trapno);
  printf("err = %lx\n", regs.err);
  printf("ip = %lx\n", regs.ip);
  printf("flags = %lx\n", regs.flags);
  printf("sp = %lx\n", regs.sp);
  printf("ss = %lx\n", regs.ss);
  printf("fs_base = %lx\n", regs.fs_base);
  printf("gs_base = %lx\n", regs.gs_base);

  return L4_EOK;
}

#define DEBUG 0

/* Handle a page fault using any configured regions. */

long ExecPager::page_fault(l4_umword_t pfa, l4_umword_t pc, l4_snd_fpage_t *region)
{
  l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

#if DEBUG
  printf("page_fault(%lx, %lx) -> %lx (%lx) -> ", pfa, pc, addr, flags);
#endif

  MappedRegions::iterator it = _regions.upper_bound(addr);

  if (it != _regions.begin())
    it--;
  else
  {
    printf("not mapped!\n");
    return -L4_ENOMEM;
  }

  MappedRegion &r = it->second;

  if ((addr >= r.map_start) && (addr < r.map_start + (1UL << r.log2size)))
  {
    l4_addr_t page_addr = trunc(addr, L4_PAGESIZE);

    region->fpage = l4_fpage(r.start + (page_addr - r.map_start), L4_PAGESHIFT, r.flags);
    region->snd_base = page_addr;

#if DEBUG
    printf("%lx...%lx from %lx...%lx size %d rights %x\n",
           r.map_start, region->snd_base,
           r.start, l4_fpage_memaddr(region->fpage),
           l4_fpage_size(region->fpage),
           l4_fpage_rights(region->fpage));
    printf("%lx -> ", addr);

    for (unsigned int i = 0; i < sizeof(l4_umword_t); i++)
      printf("%02x", *((unsigned char *)(r.start + (addr - r.map_start) + i)));

    printf("\n");
#endif

    return L4_EOK;
  }

#if DEBUG
  printf("not mapped!\n");
#endif

  return -L4_ENOMEM;
}

/* Attach a region for provision when page faults occur. This is required in
   the initialisation of a program by the C library which requires a region
   mapper. */

long ExecPager::attach(address_t *start, offset_t size, map_flags_t flags,
                       l4_cap_idx_t ds, address_t offset, unsigned char align)
{
#if DEBUG
  printf("attach(%lx, %ld, %lx, ..., %lx, %d)\n", *start, size, flags, offset, align);
#endif

  if (align < L4_PAGESHIFT)
    align = L4_PAGESHIFT;

  offset_t increment = 1UL << align;
  offset_t region_size = round(size, increment);

  /* Either attempt to find an address for the specified region, starting from
     any indicated address. */

  if (flags & L4RE_RM_F_SEARCH_ADDR)
  {
    address_t region_start = trunc(*start, increment);
    MappedRegions::iterator it = _regions.upper_bound(*start);

    if (!region_start)
      region_start += increment;

#if DEBUG
    printf("-> search from %lx -> %lx...\n", *start, region_start);
#endif

    /* Before last known region. */

    while (it != _regions.end())
    {
      MappedRegions::iterator next = it;
      MappedRegion &r = it->second;
      address_t start_limit;
      address_t end_limit = r.map_start;

      /* Consider any preceding region. If no such region exists, choose an
         address at the start of memory. */

      if (it == _regions.begin())
        start_limit = L4_PAGESIZE;
      else
      {
        it--;
        MappedRegion &pr = it->second;
        start_limit = pr.map_start + (1UL << pr.log2size);
        it = next;
      }

      /* Test against the limits. */

      if (region_start < start_limit)
        region_start = round(start_limit, increment);

      /* Investigate subsequent regions if not enough space exists between the
         preceding region (or start of memory) and the current region. */

      if ((region_start + region_size) > end_limit)
      {
        it++;
        if (it == _regions.end())
          return -L4_ENOMEM;
      }
      else
        break;
    }

    /* Attach the provided dataspace.
       NOTE: This is only done in this implementation to support the paging
             mechanism. In a region mapper residing within the actual task, the
             dataspace's map operation would be invoked to obtain mappings. */

    l4_addr_t ds_start;

    long err = ipc_attach_dataspace(ds, size, (void **) &ds_start);

    if (err)
      return err;

    l4_touch_rw((const void *) ds_start, size);

#if DEBUG
    printf("-> added region for %lx size %ld (%d)\n", region_start, region_size, page_order(region_size));
#endif

    add(MappedRegion(ds_start, page_order(region_size), flags & L4RE_DS_F_RIGHTS_MASK, region_start));

    *start = region_start;
    return L4_EOK;
  }

  /* Or attempt to add the specified region at a specific address. */

  else
  {
    // NOTE: To be implemented.

    return -L4_ENOMEM;
  }
}



/* Capability mapping definitions for the new task. */

struct mapped_cap
{
  l4_umword_t index;
  l4_cap_idx_t cap;
  unsigned char rights;
};

static long map_capabilities(l4_cap_idx_t task, struct mapped_cap mapped_caps[])
{
  long err = L4_EOK;
  int i = 0;

  while (l4_is_valid_cap(mapped_caps[i].cap) && !err)
  {
    err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                               l4_obj_fpage(mapped_caps[i].cap, 0, mapped_caps[i].rights),
                               l4_map_obj_control(mapped_caps[i].index, L4_MAP_ITEM_MAP)));
    i++;
  }

  return err;
}



/* A stack abstraction. */

class Stack
{
  struct auxv_entry
  {
    l4_umword_t key, value;
  };

protected:
  Segment &_segment;

  /* Next element pointer. */

  l4_umword_t *_element;

  /* Stack section properties. */

  l4_addr_t _caps;
  char *_arg_top, *_env_top;
  char *_auxv_end;
  int _env_entries;

public:
  /* Start address and environment structure. */

  l4_addr_t start;
  l4re_env_t *env;
  l4_fpage_t utcb_fpage;

  /* Initialise a stack in a memory segment. */

  explicit Stack(Segment &segment, l4_fpage_t utcb_fpage)
  : _segment(segment), utcb_fpage(utcb_fpage)
  {
    _element = (l4_umword_t *) (segment.buf + segment.size);
  }

  /* Push any additional initial capabilities. */

  void push_cap_entries(l4re_env_cap_entry_t *entries)
  {
    l4re_env_cap_entry_t *entry = (l4re_env_cap_entry_t *) _element;

    *(--entry) = l4re_env_cap_entry_t();

    while ((entries != NULL) && (entries->cap != L4_INVALID_CAP))
      *(--entry) = *entries;

    _caps = (l4_addr_t) entry;
    _element = (l4_umword_t *) entry;
  }

  /* Push environment values in reverse order. */

  void push_string(char *s)
  {
    char *arg = (char *) _element;
    char *arg_last = arg;

    arg -= round(strlen(s) + 1, sizeof(l4_umword_t));

    memset(arg, 0, arg_last - arg);
    memcpy(arg, s, strlen(s));

    _element = (l4_umword_t *) arg;
  }

  /* Push environment values in reverse order. */

  void push_env(char *envp[])
  {
    _env_top = (char *) _element;
    _env_entries = 0;

    for (; *envp != NULL; envp++, _env_entries++)
      push_string(*envp);
  }

  /* Push argument values in reverse order. */

  void push_args(int argc, char *argv[])
  {
    _arg_top = (char *) _element;

    for (int i = 0; i < argc; i++)
      push_string(argv[i]);
  }

  /* Loader flags, debugging flags, and the KIP capability index.
     See: generate_l4aux in Remote_app_model */

  void push_l4re_flags()
  {
    *(--_element) = 0;
    *(--_element) = 0;
    *(--_element) = 0x14 << L4_CAP_SHIFT;
  }

  void push_l4re_env()
  {
    env = (l4re_env_t *) _element;
    env--;

    env->factory          = L4_BASE_FACTORY_CAP;
    env->main_thread      = L4_BASE_THREAD_CAP;
    env->log              = L4_BASE_LOG_CAP;
    env->scheduler        = L4_BASE_SCHEDULER_CAP;
    env->rm               = 0x11 << L4_CAP_SHIFT;
    env->mem_alloc        = 0x12 << L4_CAP_SHIFT;
    env->first_free_cap   = 0x15;
    env->caps             = (l4re_env_cap_entry_t *) (_segment.region_address(_caps));
    env->utcb_area        = utcb_fpage;
    env->first_free_utcb  = l4_fpage_memaddr(utcb_fpage) + L4_UTCB_OFFSET;

    _element = (l4_umword_t *) env;
  }

  /* Push the auxiliary vector. */

  void push_auxv()
  {
    _auxv_end = (char *) _element;

    struct auxv_entry *auxv_base = (struct auxv_entry *) _element;

    /* AUXV NULL. */

    *(--auxv_base) = {0, 0};

    /* L4Re global environment pointer. */

    *(--auxv_base) = {0xf1, _segment.region_address((char *) env)};

    /* Apparently required entries.
       NOTE: The user/group identifiers should be obtained from the broader
             environment. */

    *(--auxv_base) = {AT_PAGESZ, L4_PAGESIZE};
    *(--auxv_base) = {AT_UID, 0};
    *(--auxv_base) = {AT_EUID, 0};
    *(--auxv_base) = {AT_GID, 0};
    *(--auxv_base) = {AT_EGID, 0};

    _element = (l4_umword_t *) auxv_base;
  }

  /* Fill the stack in reverse with an address, returning the size of the
     value. */

  offset_t write_address(char *arg, char **addr, char *s)
  {
    offset_t size = round(strlen(s) + 1, sizeof(l4_umword_t));

    *addr = (char *) _segment.region_address(arg - size);
    return size;
  }

  /* Populate stack with environment pointers, employing a pointer ordering
     that is the reverse of the value ordering. */

  void push_envp(char *envp[])
  {
    /* Write the terminating element. */

    *(--_element) = 0;

    /* Reserve space and fill the stack from the top inwards. */

    char **ep = (char **) (_element - _env_entries);
    char *arg = _env_top;

    for (; *envp != NULL; envp++, ep++)
      arg -= write_address(arg, ep, *envp);

    _element -= _env_entries;
  }

  /* Populate stack with argument pointers and count, employing a pointer
     ordering that is the reverse of the value ordering. */

  void push_argv(int argc, char *argv[])
  {
    /* Write the terminating element. */

    *(--_element) = 0;

    /* Reserve space and fill the stack from the top inwards. */

    char **ap = (char **) (_element - argc);
    char *arg = _arg_top;

    for (int i = 0; i < argc; i++)
      arg -= write_address(arg, &ap[i], argv[i]);

    /* Write the count. */

    _element -= argc;

    *(--_element) = argc;
  }

  /* Adjust the stack alignment and return the stack address. */

  l4_addr_t align_stack()
  {
    char *current = (char *) _element;
    char *adjusted = Ldr::adjust_sp(current, NULL);

    if (adjusted != current)
      memmove(adjusted, (const void *) current, _auxv_end - current);

    _element = (l4_umword_t *) adjusted;

    return _segment.region_address(adjusted);
  }

  /* Populate stack with additional capabilities. */

  void populate(int argc, char *argv[], char *envp[])
  {
    push_cap_entries(NULL);

    /* Populate stack with environment and argument values. */

    push_env(envp);
    push_args(argc, argv);

    /* Push L4Re flags, environment and auxiliary vector. */

    push_l4re_flags();
    push_l4re_env();
    push_auxv();

    /* Push environment and argument pointers. */

    push_envp(envp);
    push_argv(argc, argv);

    /* Adjust the stack alignment. */

    start = align_stack();
  }
};



class Environment
{
protected:
  Stack &_st;
  l4_cap_idx_t task = L4_INVALID_CAP, thread = L4_INVALID_CAP;
  l4_cap_idx_t _pager_cap = 0x10 << L4_CAP_SHIFT;

public:
  explicit Environment(Stack &st)
  : _st(st)
  {
  }

  long create_task()
  {
    task = ipc_cap_alloc();

    if (l4_is_invalid_cap(task))
      return -L4_ENOMEM;

    return l4_error(l4_factory_create_task(l4re_env()->factory, task, _st.utcb_fpage));
  }

  long create_thread()
  {
    thread = ipc_cap_alloc();

    if (l4_is_invalid_cap(thread))
      return -L4_ENOMEM;

    return l4_error(l4_factory_create_thread(l4re_env()->factory, thread));
  }

  /* Configure the task environment. */

  long configure(l4_cap_idx_t server)
  {
    /* Map the KIP into the task. */

    l4_addr_t kip_start = (l4_addr_t) l4re_kip();

    long err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                                    l4_fpage(kip_start, L4_PAGESHIFT, L4_FPAGE_RX),
                                    kip_start));

    if (err)
      return err;

    /* Define capability mappings for the new task. */

    l4re_env_t *env = _st.env;

    struct mapped_cap mapped_caps[] = {
      {L4_BASE_TASK_CAP,  task,                   L4_CAP_FPAGE_RWS},
      {_pager_cap,        server,                 L4_CAP_FPAGE_RWS},
      {env->rm,           server,                 L4_CAP_FPAGE_RWS},
      {env->main_thread,  thread,                 L4_CAP_FPAGE_RWS},
      {env->factory,      l4re_env()->factory,    L4_CAP_FPAGE_RWS},
      {env->log,          l4re_env()->log,        L4_CAP_FPAGE_RWS},
      {env->scheduler,    l4re_env()->scheduler,  L4_CAP_FPAGE_RWS},
      {env->mem_alloc,    l4re_env()->mem_alloc,  L4_CAP_FPAGE_RWS},
      {0,                 L4_INVALID_CAP,         0},
      };

    return map_capabilities(task, mapped_caps);
  }

  /* Initialise the thread. */

  long thread_init(l4_addr_t program_start)
  {
    l4_thread_control_start();
    l4_thread_control_pager(_pager_cap);
    l4_thread_control_exc_handler(_pager_cap);
    l4_thread_control_bind((l4_utcb_t *) l4_fpage_memaddr(_st.utcb_fpage), task);

    long err = l4_error(l4_thread_control_commit(thread));

    if (err)
      return err;

    return l4_error(l4_thread_ex_regs(thread, program_start, _st.start, 0));
  }

  /* Start the thread. */

  long thread_start()
  {
    l4_sched_param_t sp = l4_sched_param(L4RE_MAIN_THREAD_PRIO);

    return l4_error(l4_scheduler_run_thread(l4re_env()->scheduler, thread, &sp));
  }
};



static ExecPager exec_pager;

static void init_pager(ipc_server_config_type *config)
{
  ipc_server_init_config(config);

  config->expected_items = PagerObject_expected_items;
  config->handler = (ipc_server_handler_type) handle_PagerObject;
  config->handler_obj = static_cast<PagerObject *>(&exec_pager);
}

static long start_pager(ipc_server_config_type *config, pthread_t thread)
{
  config->config_thread = 1;
  config->thread = pthread_l4_cap(thread);

  printf("Starting pager thread...\n");
  return ipc_server_start_config(config);
}



int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need a program to run.\n");
    return 1;
  }

  /* Obtain the payload as a dataspace. */

  file_t *file = client_open(argv[1], O_RDONLY);

  if (file == NULL)
  {
    printf("Could not read file: %s\n", argv[1]);
    return 1;
  }

  /* Copy the payload regions to new dataspaces.
     NOTE: This should be directed by the ELF metadata. */

  address_t program_start = 0x1000b43;
  offset_t initial_stack_size = 16 * L4_PAGESIZE;

  Segment program(0x1000000, 0x28326, L4_FPAGE_RWX, 0, 0x28326);
  Segment data(0x1029360, 0x8068, L4_FPAGE_RW, 0x28360, 0x2058);
  Segment stack(Utcb_area_start - initial_stack_size, initial_stack_size, L4_FPAGE_RW);
  Segment *segments[] = {&program, &data, &stack, NULL};
  Segment **segment;

  for (segment = segments; *segment != NULL; segment++)
  {
    err = (*segment)->allocate();

    if (err)
    {
      printf("Could not reserve memory.\n");
      return 1;
    }

    err = (*segment)->fill(file);

    if (err)
    {
      printf("Could not fill segment from file.\n");
      return 1;
    }
  }

  /* Populate the stack with argument and environment details. */

  Stack st(stack, get_utcb_fpage());
  Environment environment(st);

  /* NOTE: Environment vector is currently not defined. */

  char *envp[] = {NULL};

  st.populate(argc - 1, argv + 1, envp);

  /* Create a new task and thread. */

  err = environment.create_task();

  if (err)
  {
    printf("Could not create task.\n");
    return 1;
  }

  err = environment.create_thread();

  if (err)
  {
    printf("Could not create thread.\n");
    return 1;
  }

  err = environment.thread_init(program_start);

  if (err)
  {
    printf("Could not configure thread.\n");
    return 1;
  }

  /* Start the pager. */

  ipc_server_config_type config;
  pthread_t pager_thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  init_pager(&config);

  for (segment = segments; *segment != NULL; segment++)
    exec_pager.add((*segment)->region());

  err = pthread_create(&pager_thread, &attr, ipc_server_start_mainloop, &config);

  if (err)
  {
    printf("Could not start pager thread.\n");
    return 1;
  }

  err = start_pager(&config, pager_thread);

  if (err)
  {
    printf("Could not start pager.\n");
    return 1;
  }

  /* Configure the environment for the task, specifying the pager (and exception
     handler plus region mapper). */

  err = environment.configure(config.server);

  if (err)
  {
    printf("Could not map capabilities into task.\n");
    return 1;
  }

  /* Start the new thread. */

  printf("Run thread...\n");

  err = environment.thread_start();

  if (err)
  {
    printf("Could not run thread.\n");
    return 1;
  }

  printf("Finished.\n");
  while (1);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
